#include "stdafx.h"
#include "JavaScriptSyntaxColorer.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

JavaScriptSyntaxColorState::JavaScriptSyntaxColorState()
{
	state = 0;
}

JavaScriptSyntaxColorState::JavaScriptSyntaxColorState(JavaScriptSyntaxColorState* original)
{
	stateStack = original->stateStack;
	state = original->state;
}

//	Java keywords
static LPTSTR s_apszJavaScriptKeywordList[] =
{
	_T("abstract"),
	_T("boolean"),
	_T("break"),
	_T("byte"),
	_T("case"),
	_T("catch"),
	_T("char"),
	_T("class"),
	_T("const"),
	_T("continue"),
	_T("debugger"),
	_T("default"),
	_T("delete"),
	_T("do"),
	_T("double"),
	_T("else"),
	_T("enum"),
	_T("export"),
	_T("extends"),
	_T("false"),
	_T("final"),
	_T("finally"),
	_T("float"),
	_T("for"),
	_T("function"),
	_T("goto"),
	_T("if"),
	_T("implements"),
	_T("import"),
	_T("in"),
	_T("instanceof"),
	_T("int"),
	_T("interface"),
	_T("long"),
	_T("native"),
	_T("new"),
	_T("null"),
	_T("package"),
	_T("private"),
	_T("protected"),
	_T("public"),
	_T("return"),
	_T("short"),
	_T("static"),
	_T("super"),
	_T("switch"),
	_T("synchronized"),
	_T("this"),
	_T("throw"),
	_T("throws"),
	_T("transient"),
	_T("true"),
	_T("try"),
	_T("typeof"),
	_T("var"),
	_T("void"),
	_T("volatile"),
	_T("while"),
	_T("with"),
	NULL
};

static BOOL IsJavaScriptKeyword(LPCTSTR pszChars, int nLength)
{
	for (int L = 0; s_apszJavaScriptKeywordList[L] != NULL; L ++)
	{
		if (strncmp(s_apszJavaScriptKeywordList[L], pszChars, nLength) == 0
				&& s_apszJavaScriptKeywordList[L][nLength] == 0)
			return TRUE;
	}
	return FALSE;
}

static BOOL IsJavaScriptNumber(LPCTSTR pszChars, int nLength)
{
	if (nLength > 2 && pszChars[0] == '0' && pszChars[1] == 'x')
	{
		for (int I = 2; I < nLength; I++)
		{
			if (isdigit(pszChars[I]) || (pszChars[I] >= 'A' && pszChars[I] <= 'F') ||
										(pszChars[I] >= 'a' && pszChars[I] <= 'f'))
				continue;
			return FALSE;
		}
		return TRUE;
	}
	if (! isdigit(pszChars[0]))
		return FALSE;
	for (int I = 1; I < nLength; I++)
	{
		if (! isdigit(pszChars[I]) && pszChars[I] != '+' &&
			pszChars[I] != '-' && pszChars[I] != '.' && pszChars[I] != 'e' &&
			pszChars[I] != 'E')
			return FALSE;
	}
	return TRUE;
}

#define COOKIE_COMMENT			0x0001
#define COOKIE_EXT_COMMENT		0x0002
#define COOKIE_STRING			0x0004
#define COOKIE_CHAR				0x0008

#define COLOR_WHITESPACE	RGB(255, 255, 255)
#define COLOR_NORMALTEXT	RGB(  0,   0,   0)
#define COLOR_COMMENT		RGB( 64,  64, 220)
#define COLOR_NUMBER		RGB(128,   0,   0)
#define COLOR_OPERATOR		RGB(  0,   0,   0)
#define COLOR_STRING		RGB(  0, 128,   0)
#define COLOR_BK_STRING		RGB(240, 255, 240)
#define COLOR_KEYWORD		RGB(  0,  64, 128)
#define COLOR_SELBKGND		RGB(  0,   0,   0)
#define COLOR_SELTEXT		RGB(255, 255, 255)
#define COLOR_TEXT_ERROR	RGB(128, 0, 0)
#define COLOR_BK_ERROR		RGB(255, 240, 240)

JavaScriptSyntaxColorer::JavaScriptSyntaxColorer()
{
	defaultTextColor = ::GetSysColor(COLOR_WINDOWTEXT);
	defaultBkColor = ::GetSysColor(COLOR_WINDOW);
	noTextColor = defaultBkColor;
	selectTextColor = RGB(255, 255, 255);
	selectBkColor = RGB(0, 0, 0);
}

void JavaScriptSyntaxColorer::parseLine(const char* line, JavaScriptSyntaxColorState &JavaScriptstate, int &lineFlags)
{
	SharedStack<char> *stateStack = &JavaScriptstate.stateStack;
	int state = JavaScriptstate.state;

	BOOL bFirstChar     = (state & ~COOKIE_EXT_COMMENT) == 0;
	BOOL bRedefineBlock = TRUE;
	BOOL bDecIndex  = FALSE;
	int nIdentBegin = -1;
	for (int I = 0; ; I++)
	{
		if (bRedefineBlock)
		{
			int nPos = I;
			if (bDecIndex)
				nPos--;
			if (state & (COOKIE_COMMENT | COOKIE_EXT_COMMENT))
			{
				startColor(nPos, COLOR_COMMENT, COLOR_WHITESPACE);
			}
			else
			if (state & (COOKIE_CHAR | COOKIE_STRING))
			{
				startColor(nPos, COLOR_STRING, COLOR_BK_STRING);
			}
			else
			{
				startColor(nPos, COLOR_NORMALTEXT, COLOR_WHITESPACE);
			}
			bRedefineBlock = FALSE;
			bDecIndex      = FALSE;
		}

		if (line[I] == '\0')
			break;

		if (state & COOKIE_COMMENT)
		{
			startColor(I, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	String constant "...."
		if (state & COOKIE_STRING)
		{
			if (line[I] == '"' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_STRING;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Char constant '..'
		if (state & COOKIE_CHAR)
		{
			if (line[I] == '\'' && (I == 0 || line[I - 1] != '\\' || (I >= 2 && line[I - 2] == '\\')))
			{
				state &= ~COOKIE_CHAR;
				bRedefineBlock = TRUE;
				endColor(I);
			}
			continue;
		}

		//	Extended comment /*....*/
		if (state & COOKIE_EXT_COMMENT)
		{
			if (I > 0 && line[I] == '/' && line[I - 1] == '*')
			{
				state &= ~COOKIE_EXT_COMMENT;
				bRedefineBlock = TRUE;
			}
			continue;
		}

		if (I > 0 && line[I] == '/' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_COMMENT;
			break;
		}

		//	Normal text
		if (line[I] == '"')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_STRING;
			continue;
		}
		if (line[I] == '\'')
		{
			startColor(I+1, COLOR_STRING, COLOR_BK_STRING);
			state |= COOKIE_CHAR;
			continue;
		}
		if (I > 0 && line[I] == '*' && line[I - 1] == '/')
		{
			startColor(I - 1, COLOR_COMMENT, COLOR_WHITESPACE);
			state |= COOKIE_EXT_COMMENT;
			continue;
		}

		if (nIdentBegin >= 0 && !(isalnum(line[I]) || line[I] == '_' || line[I] == '.'))
		{
			if (IsJavaScriptKeyword(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
			}
			else if (IsJavaScriptNumber(line + nIdentBegin, I - nIdentBegin))
			{
				startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
			}
			bRedefineBlock = TRUE;
			bDecIndex = TRUE;
			nIdentBegin = -1;
		}

		if (line[I] == '(')
		{	
			stateStack->push(')');
		}
		else if (line[I] == ')')
		{	
			if (stateStack->topIs(')'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '[')
		{	
			stateStack->push(']');
		}
		else if (line[I] == ']')
		{	if (stateStack->topIs(']'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (line[I] == '{')
		{	
			if (stateStack->topIs(')') || stateStack->topIs(']'))
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
			stateStack->push('}');
		}
		else if (line[I] == '}')
		{	
			if (stateStack->topIs('}'))
				stateStack->pop();
			else
				startColor(I, COLOR_TEXT_ERROR, COLOR_BK_ERROR);
		}
		else if (isalnum(line[I]) || line[I] == '_' || line[I] == '.')
		{
			if (nIdentBegin == -1)
				nIdentBegin = I;
		}
		else
		{
		}
	}

	if (nIdentBegin >= 0)
	{
		if (IsJavaScriptKeyword(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_KEYWORD, COLOR_WHITESPACE);
		}
		else
		if (IsJavaScriptNumber(line + nIdentBegin, I - nIdentBegin))
		{
			startColor(nIdentBegin, COLOR_NUMBER, COLOR_WHITESPACE);
		}
	}

	if (I > 0 && line[I - 1] != '\\')
		state &= COOKIE_EXT_COMMENT;

	JavaScriptstate.state = state;
}
